/*
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <time.h>

#include "libtkcommon/string.h"
#include "libtkcommon/intl.h"
#include "libtkrandom/randomness.h"
#include "libtkrandom/retriever_registry.h"
#include "libtkrandom/rng.h"
#include "libtkdb/value.h"
#include "experimenter/experimenter.h"
#include "experimenter/text_frame.h"
#include "experimenter/catcher.h"
#include "experimenter/variables.h"
#include "experimenter/database_operation.h"

using namespace Tk::Experimenter;

Experimenter:: Experimenter(Db::Database *database)
  : _database(database),
    _control(Gtk::Stock::MEDIA_RECORD, _toolbar.get_icon_size(), Gdk::Color("Blue")),
    _results_store(*this, database),
    _parameters(database),
    _location(database),
    _operators(database),
    _notes(Gtk::TextBuffer::create())
    
{
  WDEBUG_OUT << L"Initialising Experimenter..." << std::endl;
  
  set_title("TkResolver++ Experimenter");
  set_default_size(980, 500);
  set_position(Gtk::WIN_POS_CENTER);

  sigc::slot<void> execute_slot( sigc::mem_fun(*this, &Experimenter::execute) );
  sigc::slot<void> test_slot   ( sigc::mem_fun(*this, &Experimenter::test) );
  sigc::slot<void> control_slot( sigc::mem_fun(*this, &Experimenter::control_toggle) );
  sigc::slot<void> quit_slot   ( sigc::ptr_fun(&Gtk::Main::quit) );

  Gtk::VBox *box = Gtk::manage(new Gtk::VBox);
  add(*box);

  // toolbar
  box->pack_start(_toolbar, false, true);

  Gtk::ToolButton *tool_execute = Gtk::manage(new Gtk::ToolButton(Gtk::Stock::EXECUTE));
  _toolbar.append(*tool_execute, execute_slot);

  _toolbar.append(_control, control_slot);
  _toolbar.append(_results_store);

  Gtk::ToolButton *tool_quit = Gtk::manage(new Gtk::ToolButton(Gtk::Stock::QUIT));
  _toolbar.append(*tool_quit, quit_slot);
  
  // toolbar tooltips
  /*
  Gtk::Tooltips *tooltips = _toolbar.get_tooltips_object();
  tooltips->set_tip(*tool_execute, "Execute single experiment");
  tooltips->set_tip(_control,      "Start control experiments");
  tooltips->set_tip(*tool_quit,    "Exit the experimenter");
  tooltips->enable();
  _toolbar.set_tooltips(true);
  */
  tool_execute->set_tooltip_text(_("Execute single experiment"));
  _control.set_tooltip_text(_("Start control experiments"));
  _results_store.set_tooltip_text(_("Store results before writing to database (click to flush)"));
  tool_execute->set_tooltip_text(_("Exit the experimenter"));
  
  /// main controls
  Gtk::HPaned *controls( Gtk::manage(new Gtk::HPaned) );
  box->pack_start(*controls, true, true);
  
  // parameters
  controls->pack1(_parameters, Gtk::EXPAND|Gtk::FILL);
  
  // location, operators and notes
  Gtk::VPaned *lists_and_notes = Gtk::manage(new Gtk::VPaned);
  controls->pack2(*lists_and_notes, Gtk::EXPAND|Gtk::FILL);
  
  Gtk::VPaned *lists = Gtk::manage(new Gtk::VPaned);
  lists_and_notes->pack1(*lists);
  lists->property_position() = 120;
  
  lists->pack1(_location,                 Gtk::FILL);
  lists->pack2(_operators, Gtk::EXPAND|Gtk::FILL);

  TextFrame *notes_frame = Gtk::manage(new TextFrame("Notes", _notes));
  lists_and_notes->pack2(*notes_frame, Gtk::FILL);
  
  // bottom execute buttons
  Gtk::HBox *exec_box = Gtk::manage(new Gtk::HBox());
  box->pack_start(*exec_box, false, true);
  
  Gtk::Button *test = Gtk::manage(new Gtk::Button("Test"));
  exec_box->pack_start(*test, false, true);
  test->signal_clicked().connect(test_slot);
  
  Gtk::Button *execute = Gtk::manage(new Gtk::Button(Gtk::Stock::EXECUTE));
  exec_box->pack_start(*execute, true, true);
  execute->signal_clicked().connect(execute_slot);

  show_all_children();
  show();
  
  WDEBUG_OUT << "...initialised Experimenter" << std::endl;
}


void
Experimenter:: execute_proper(bool test,
			      const Db::Ids & operators,
			      bool control,
			      const String * notes,
			      bool interactive)
{
  //
  // get parameters
  //
  Db::Id location_id;
  if (!get_location_id(location_id))
    return;

  Tk::String retriever_type( _parameters.get_retriever_type() );
  Tk::String algorithm_spec( _parameters.get_algorithm_spec() );
  unsigned bit_multiplier( _parameters.get_bit_multiplier() );
  bool coloured( _parameters.get_coloured() );
  unsigned height(_parameters.get_height());
  unsigned width(_parameters.get_width());
  unsigned trace_count( _parameters.get_trace_count() );
  unsigned total_time( _parameters.get_total_time() );
  double horizon( _parameters.get_horizon() );
  
  // establish intent and ability rating
  Tk::StringValue intent;
  Double ability_rating;
  if (interactive)
    {
      Variables *variables = new Variables(_database, *this, trace_count, algorithm_spec, test);
      
      int response = variables->run();
      if (response != Gtk::RESPONSE_OK)
	{
	  delete variables;
	  return;
	}
      
      intent = variables->get_intent_code();
      ability_rating = variables->get_ability_rating();
      delete variables;
    }
  
  Tk::Resolver::DirectionValue arrow;
  if (intent == Tk::String(L"LT"))
    arrow = Tk::Resolver::Left;
  else if (intent == Tk::String(L"RT"))
    arrow = Tk::Resolver::Right;
  
  // do it
  Tk::Random::RetrieverPtr retriever( Tk::Random::RetrieverRegistry::new_retriever(retriever_type) );
  Tk::Random::Randomness randomness(true);
  Tk::Resolver::Resolver resolver(algorithm_spec, bit_multiplier);
  
  unsigned retries;
  Tk::Resolver::Time actual_time;
  Tk::Resolver::Result *result;
  
  Tk::Db::DateTime start_date_time;
  try
    {
      result = resolver.run(*retriever,
			    randomness,
			    coloured,
			    height,
			    width,
			    trace_count,
			    arrow,
			    total_time,
			    horizon,
			    test,
			    retries,
			    actual_time);
    }
  catch (Tk::Resolver::Escaped)
    {
      return;
    }
	       
  using namespace Tk::Db;
  using namespace Tk::Db::Data;
  ValueMap *props( new ValueMap );
  props->insert(ValueMap::value_type("time stamp", start_date_time));
  props->insert(ValueMap::value_type("test", test));
  props->insert(ValueMap::value_type("algorithm spec", algorithm_spec));
  props->insert(ValueMap::value_type("bit multiplier", static_cast<long>(bit_multiplier)));
  props->insert(ValueMap::value_type("coloured", coloured));
  props->insert(ValueMap::value_type("trace count", static_cast<long>(trace_count)));
  props->insert(ValueMap::value_type("time", static_cast<long>(total_time)));
  props->insert(ValueMap::value_type("actual time", static_cast<long>(actual_time)));
  props->insert(ValueMap::value_type("horizon", horizon));
  props->insert(ValueMap::value_type("notes", Data::Value(notes ? *notes : static_cast<Tk::String>(_notes->get_text()))));
  props->insert(ValueMap::value_type("intent code", static_cast<Data::Value>(intent)));
  props->insert(ValueMap::value_type("ability rating", static_cast<Data::Value>(ability_rating)));
  props->insert(ValueMap::value_type("screen width", static_cast<long>(width)));
  props->insert(ValueMap::value_type("screen height", static_cast<long>(height)));
  props->insert(ValueMap::value_type("retriever name", retriever_type));
  props->insert(ValueMap::value_type("control", control));

  const Tk::Random::Bits *used_bits( randomness.get_used_bits() );
  props->insert(ValueMap::value_type("randomness bits", static_cast<long>(used_bits->size())));
  
  Tk::Random::BitStream *used_bits_stream( used_bits->get_data() );
  Bytes used_bits_bytes(*used_bits_stream->get_stream());
  delete used_bits_stream;
  props->insert(ValueMap::value_type("randomness", used_bits_bytes));
  
  
  Tk::Db::ExperimentResult *experiment( new Tk::Db::ExperimentResult(location_id,
								     operators,
								     props,
								     result) );
  _results_store.store_results(experiment);
}

bool
Experimenter:: get_location_id(Db::Id &location_id, bool interactive)
{
  try
    {
      location_id = _location.get_location();
    }
  catch (const Db::Ui::NoSelection & ns)
    {
      if (interactive)
	{
	  Gtk::MessageDialog dialog(*this, "No location selected", false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
	  dialog.set_title("Execution error");
	  dialog.run();
	}
      return false;
    }

  return true;
}


void
Experimenter:: execute_or_test(bool test)
{
  Db::Ids operators;
  try
    {
      operators = _operators.get_operators();
    }
  catch (const Db::Ui::NoSelection & ns)
    {
      Gtk::MessageDialog dialog(*this, "No operators selected", false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
      dialog.set_title("Execution error");
      dialog.run();
      return;
    }

  execute_proper(test, operators, false);
}

void
Experimenter:: execute()
{
  execute_or_test(false);
}

void
Experimenter:: test()
{
  try
    {
      execute_or_test(true);
    }
  catch (std::runtime_error e)
    {
      std::cerr << "Caught exception ('" << e.what() << "') while executing test" << std::endl;
      Tk::Resolver::SDL::reset();
    }
}

void
Experimenter:: control_toggle()
{
  if (_control.get_active())
    {
      if (!control_start())
	_control.set_active(false);
    }
  else
    control_stop();
}

bool
Experimenter:: control_start()
{
  Db::Id location_id; // we're not going to use this; just checking
  if (!get_location_id(location_id, false))
    return false;
  
  String retriever_type( _parameters.get_retriever_type() );

  // anything between 10 minutes to 1 hour
  const unsigned lower_wait = 1000 * 60 * 10;
  const unsigned upper_wait = 1000 * 60 * 60;
  //  const unsigned lower_wait = 1000 * 2;
  //  const unsigned upper_wait = 1000 * 4;
  unsigned wait_length(lower_wait + Tk::Random::random<unsigned>(lower_wait, upper_wait, retriever_type));
  WDEBUG_OUT << L"control waiting for " << wait_length << L" milliseconds (" << (wait_length/1000/60) << " minutes)" << std::endl;
  _timeout = Glib::signal_timeout().connect(sigc::mem_fun(*this, &Experimenter::control_notify), wait_length);

  return true;
}

bool
Experimenter:: control_notify()
{
  _control.flash_start(500);
  unsigned notify_time = 1000 * 30;
  //  unsigned notify_time = 1000 * 10;
  _timeout = Glib::signal_timeout().connect(sigc::mem_fun(*this, &Experimenter::control_exec), notify_time);
  return false;
}

void
Experimenter:: control_stop()
{
  _timeout.disconnect();
  _control.flash_stop();
}

bool
Experimenter:: control_exec()
{
  DEBUG_OUT << "execing control" << std::endl;
  control_stop();
  
  Db::Ids null_operators;  
  
  String notes("Control");
  
  String user_notes(_notes->get_text());
  if (user_notes.size() > 0)
    {
      notes += L" (";
      notes += user_notes;
      notes += L")";
    }
  
  execute_proper(false, null_operators, true, &notes, false);
  
  control_start();
  return false;
}
