/* -*- c++ -*-
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2009 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef __TK_STREAM_H__
#define __TK_STREAM_H__

#include <sstream>
#include <limits>

#include "libtkcommon/string.h"
#include "libtkcommon/debug.h"

namespace Tk {
  
  template <typename C>
  class BasicStream
  {
  public:
    typedef std::basic_istream<C> istream_type;
    typedef std::basic_stringstream<C> stringstream_type;
    typedef std::basic_istringstream<C> istringstream_type;
    typedef std::basic_string<C> string_type;
    typedef typename string_type::size_type len_type;
    typedef typename string_type::value_type char_type;
  private:
    istream_type *_stream;
    len_type      _len;
  public:
    BasicStream();
    BasicStream(istream_type &src_stream, len_type src_len = -1);
    BasicStream(const C *data, len_type len);
    template <typename O>
    BasicStream(std::basic_istream<O> &src_stream, typename std::basic_string<O>::size_type src_len = -1);
    
    virtual ~BasicStream()
    { delete _stream; }
  
    istream_type *get_stream()
    { return _stream; }
    const istream_type *get_stream() const
    { return _stream; }
    len_type get_len() const
    { return _len; }
    
    BasicStream<C> *write_stream();
    template <typename O>
    BasicStream<O> *write_stream();
    
  private:
    static const unsigned short BITS_PER_CHAR = std::numeric_limits<C>::digits + static_cast<unsigned>(std::numeric_limits<C>::is_signed);
  };

  typedef BasicStream<char> Stream;
  
  
  template <typename C>
  BasicStream<C>:: BasicStream()
  : _stream(new istringstream_type), _len(0) {}
  
  template <typename C>
  BasicStream<C>:: BasicStream(istream_type &src_stream, len_type src_len)
  : _stream(0),
    _len(0)
  {
#ifdef TK_STREAM_DEBUG
    DEBUG_OUT << "Bits per char: " << BITS_PER_CHAR << " (no conversion)" << std::endl;
#endif
    stringstream_type *dest( new stringstream_type );
    C src_char;
    
    for (len_type src_read(0);
	 (src_len == -1 || src_read < src_len) && src_stream.read(&src_char, 1);
	 ++src_read)
      {
	dest->put(src_char);
	++_len;
      }
    
#ifdef TK_STREAM_DEBUG
    DEBUG_OUT << "Length: " << _len << " (" << (_len * BITS_PER_CHAR) << " bits)" << std::endl;
#endif
    _stream = dest;
  }
  
  template <typename C> template <typename O>
  BasicStream<C>:: BasicStream(std::basic_istream<O> &src_stream, typename std::basic_string<O>::size_type src_len)
  : _stream(0),
    _len(0)
  {
    static const unsigned short BITS_PER_SRC_CHAR = std::numeric_limits<O>::digits + static_cast<unsigned>(std::numeric_limits<O>::is_signed);
    stringstream_type *dest( new stringstream_type );
    O src_char;
    C dest_char(0);
    unsigned short src_char_index(0);
    unsigned short dest_char_index(0);
      
#ifdef TK_STREAM_DEBUG
    DEBUG_OUT << "Bits per char: `"
	      << BITS_PER_CHAR 
	      << "' ("
	      << (std::numeric_limits<C>::is_signed ? "signed" : "unsigned" )
	      <<"); bits per source char: `"
	      << BITS_PER_SRC_CHAR
	      << "' ("
	      << (std::numeric_limits<O>::is_signed ? "signed" : "unsigned" )
	      << ")" << std::endl;
#endif
    
    dest->exceptions(std::ios_base::badbit);
    
    typename std::basic_string<O>::size_type src_read;
    for (src_read = 0;
	 (src_len == -1 || src_read < src_len) && src_stream.read(&src_char, 1);
	 ++src_read)
      {
#ifdef TK_STREAM_DEBUG
	{
	  std::ostringstream ostrm;
	  for (unsigned i = 0; i < BITS_PER_SRC_CHAR; ++i)
	    ostrm << ((src_char & (1 << (BITS_PER_SRC_CHAR - i - 1))) ? "1" : "0");
	  DEBUG_OUT << "Read char " << src_read << ":\t`" << ostrm.str() << "' (`" << src_char << "')"<< std::endl;
	}
#endif
	for (src_char_index = 0; src_char_index < BITS_PER_SRC_CHAR; ++src_char_index)
	  {
	    bool bit( src_char & (1 << (BITS_PER_SRC_CHAR - src_char_index - 1)) );
	    
	    dest_char |= static_cast<unsigned>(bit) << (BITS_PER_CHAR - dest_char_index - 1);
	    
	    if (dest_char_index == BITS_PER_CHAR - 1)
	      {
#ifdef TK_STREAM_DEBUG
		{
		  std::ostringstream ostrm;
		  for (unsigned i = 0; i < BITS_PER_CHAR; ++i)
		    ostrm << ((dest_char & (1 << (BITS_PER_CHAR - i - 1))) ? "1" : "0");
		  DEBUG_OUT << "Dest char " << _len << ":\t`" << ostrm.str() << "' (`" << dest_char << "')" << std::endl;
		}
#endif

		dest->write(&dest_char, 1);
		//dest->put(dest_char);
		++_len;
		dest_char = 0;
		dest_char_index = 0;
	      }
	    else
	      ++dest_char_index;
	  }
      }
      
    if (dest_char_index > 0)
      {
#ifdef TK_STREAM_DEBUG
	{
	  std::ostringstream ostrm;
	  for (unsigned i = 0; i < BITS_PER_CHAR; ++i)
	    ostrm << ((dest_char & (1 << (BITS_PER_CHAR - i - 1))) ? "1" : "0");
	  DEBUG_OUT << "Dest char " << _len << ":\t`" << ostrm.str() << "' (`" << dest_char << "')" << std::endl;
	}
#endif
	dest->write(&dest_char, 1);
	//dest->put(dest_char);
	++_len;
      }
    
    // Ensure that reading works (apparently..)
    dest->seekg(std::ios::beg);
    
    _stream = dest;

#ifdef TK_STREAM_DEBUG
    DEBUG_OUT << "Total characters read:    " << src_read << " (" << (src_read * BITS_PER_SRC_CHAR) << " bits)" << std::endl;
    DEBUG_OUT << "Total characters written: " << _len << " (" << (_len * BITS_PER_CHAR) << " bits)" << std::endl;
#endif
  }
  
  template <typename C>
  BasicStream<C>:: BasicStream(const C *data, len_type len)
    : _stream( new istringstream_type(string_type(data, len)) ), _len(len) {}
  
  template <typename C>
  BasicStream<C> *
  BasicStream<C>:: write_stream()
  {
    return new BasicStream<C>(*_stream);
  }
  
  template <typename C> template <typename O>
  BasicStream<O> *
  BasicStream<C>:: write_stream()
  {
    return new BasicStream<O>(*_stream);
  }
  
}; /* namespace Tk */


#endif /* __TK_STREAM_H__ */
