/* -*- c++ -*-
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef __TK_STRING_H__
#define __TK_STRING_H__

#include <wchar.h>

#include <string>
#include <sstream>
#include <memory>
#include <cstring>

#include <glibmm.h>
#include <gmpxx.h>

#include "libtkcommon/mbs.h"
#include "libtkcommon/debug.h"
#include "libtkcommon/numerical-types.h"

namespace Tk {

  class String : public std::wstring
    {
    public:
      String() {}
      String(const wchar_t * wstr)
	: std::wstring(wstr) {}
      String(const char * str)
	: std::wstring(mbs_to_wstring(str)) {}
      String(const std::wstring & wstr)
	: std::wstring(wstr) {}
      String(const std::string & sstr)
	: std::wstring(string_to_wstring(sstr)) {}
      String(const Glib::ustring & ustr)
	: std::wstring(ustring_to_wstring(ustr)) {} 
      
      String(bool b)               { std::wostringstream o; o << std::boolalpha << b; *this = o.str(); }
      String(char c)               { std::ostringstream  o; o << c; *this = o.str(); }
      String(wchar_t c)            { std::wostringstream o; o << c; *this = o.str(); }
      String(int i)                { std::wostringstream o; o << i; *this = o.str(); }
      String(unsigned i)           { std::wostringstream o; o << i; *this = o.str(); }
      String(long i)               { std::wostringstream o; o << i; *this = o.str(); }
      String(unsigned long i)      { std::wostringstream o; o << i; *this = o.str(); }
      String(long long i)          { std::wostringstream o; o << i; *this = o.str(); }
      String(unsigned long long i) { std::wostringstream o; o << i; *this = o.str(); }
      String(float f)              { std::wostringstream o; o << std::fixed << f; *this = o.str(); }
      String(double d)             { std::wostringstream o; o << std::fixed << d; *this = o.str(); }
      String(long double d)        { std::wostringstream o; o << std::fixed << d; *this = o.str(); }
      
      String(const mpz_class &z)
	{ std::ostringstream o; o << z; *this = o.str(); }
      String(const mpf_class &f, unsigned long precision = 72)
	{ std::ostringstream o; o.precision(precision); o << f; *this = o.str(); }
      
      /** Stringify a rational into the form
       *
       *    1.1111111111111 (222222222222222222222222222222222222/33333333333...)
       */
      String(const mpq_class &q, unsigned long precision = 30, size_type max_len = 72);
      

      String & operator= (const wchar_t * wstr)
	{ return *this = String(wstr); }
      String & operator= (const char * str)
	{ return *this = String(str); }
      String & operator= (const std::string & sstr)
	{ return *this = string_to_wstring(sstr); }
      String & operator= (const Glib::ustring & ustr)
        { return *this = ustring_to_wstring(ustr); }
      
      operator std::string() const
	{ return wstring_to_string(*this); }
      operator Glib::ustring() const
        { return wstring_to_ustring(*this); }
    };
  
  typedef std::auto_ptr<String> StringPtr;
  
  class Strings : public std::list<String>
    {
    public:
      Strings() {}
      template <class InputIterator>
      Strings(InputIterator f, InputIterator l)
	: std::list<String>(f, l) {}
      Strings(iterator &start, iterator &end)
	: std::list<String>(start, end) {}
      Strings merge(const Strings &list) const;
      Strings uniq() const;
    };
  
  typedef std::auto_ptr<Strings> StringsPtr;

}; /* namespace Tk */


/* Global operators */

/* + */
/*static Tk::String operator+(const Tk::String &tkstr1, const Tk::String &tkstr2)
{ return static_cast<std::wstring>(tkstr1) + static_cast<std::wstring>(tkstr2); }
static std::string operator+(const std::string &str, const Tk::String &tkstr)
{ return str + static_cast<std::string>(tkstr); }
static Tk::String operator+(const Tk::String &tkstr, const std::string &str)
{ return tkstr + static_cast<Tk::String>(str); }
static Tk::String operator+(const Tk::String &tkstr, const char *str)
{ return tkstr + static_cast<Tk::String>(str); }*/

/* += */
/*static Tk::String operator+=(const Tk::String &tkstr1, const Tk::String &tkstr2)
{ return static_cast<std::wstring>(tkstr1) += static_cast<std::wstring>(tkstr2); }*/
/*static Tk::String operator+=(const Tk::String &tkstr, const std::string &str)
{ return tkstr += static_cast<Tk::String>(str); }
static Tk::String operator+=(const Tk::String &tkstr, const char *str)
{ return tkstr += static_cast<Tk::String>(str); }*/

/* == */
static bool operator==(const Tk::String &tkstr1, const Tk::String &tkstr2)
{ return static_cast<std::wstring>(tkstr1) == static_cast<std::wstring>(tkstr2); }
static bool operator==(const std::string &str, const Tk::String &tkstr)
{ return static_cast<Tk::String>(str) == tkstr; }
static bool operator==(const char *str, const Tk::String &tkstr)
{ return static_cast<Tk::String>(str) == tkstr; }

/* << */


#endif /* __TK_STRING_H__ */
