/*
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <time.h>
#include <pthread.h>

#include <sstream>
#include <algorithm>

#include <boost/variant/get.hpp>
#include <boost/variant/static_visitor.hpp>
#include <boost/variant/apply_visitor.hpp>

#include <odbc++/setup.h>
#include <odbc++/drivermanager.h>
#include <odbc++/databasemetadata.h>

#include "libtkcommon/intl.h"
#include "libtkcommon/mbs.h"
#include "libtkcommon/exceptions.h"
#include "libtkcommon/value.h"
#include "libtkcommon/debug.h"
#include "libtkdb/database.h"
#include "libtkdb/value.h"
#include "libtkdb/exception.h"
#include "libtkquery/query.h"

using namespace Tk::Db;




Database:: Database (const Tk::String &data_source_name,
		     const Tk::String &username,
		     const Tk::String &password)
  : _connection(data_source_name, username, password)
{
  _report_dispatcher.connect( sigc::mem_fun(*this, &Database::postpone_reporting) );
}

Connection *
Database:: get_connection()
{
  return &_connection;
}

void
Database:: send_report(Report *report)
{
  /*
  WDEBUG_OUT << L"Sending report with description `"
	     << report->get_description()
	     << L"'"
	     << std::endl;
  */
  {
    //WDEBUG_OUT << L"Locking reporting lock (thread: " << pthread_self() << L")" << std::endl;
    Connection::Mutex::Lock lock( _reporting_lock );
    //WDEBUG_OUT << L"Locked reporting lock (thread: " << pthread_self() << L")" << std::endl;
    
    _running_reports.push_back(report);
    _pending_reports.push_back(report);
  }
  _report_dispatcher.emit();
  //WDEBUG_OUT << L"Unlocked reporting lock (thread: " << pthread_self() << L")" << std::endl;;
}

void
Database:: send_report_update(ReportUpdate *update)
{
  /*
  WDEBUG_OUT << L"Sending report update with fraction of `"
	     << update->get_fraction()
	     << L"' for report with description `"
	     << update->get_report().get_description()
	     << L"'"
	     << std::endl;
  */
  {
    //WDEBUG_OUT << L"Locking reporting lock (thread: " << pthread_self() << L")" << std::endl;
    Connection::Mutex::Lock lock( _reporting_lock );
    //WDEBUG_OUT << L"Locked reporting lock (thread: " << pthread_self() << L")" << std::endl;
    _report_updates.push_back(update);
  }
  _report_dispatcher.emit();
  //WDEBUG_OUT << L"Unlocked reporting lock (thread: " << pthread_self() << L")" << std::endl;;
}

void
Database:: postpone_reporting()
{
  Glib::signal_idle().connect( sigc::mem_fun(*this, &Database::receive_reporting) );
}

bool
Database:: receive_reporting()
{
  Glib::Thread::yield();
  Reports pending;
  ReportUpdates updates;
  {
    {
      //WDEBUG_OUT << L"Locking reporting lock (thread: " << pthread_self() << L")" << std::endl;
      Connection::Mutex::Lock lock( _reporting_lock );
      //WDEBUG_OUT << L"Locked reporting lock (thread: " << pthread_self() << L")" << std::endl;
      
      pending = _pending_reports;
      _pending_reports.clear();
      
      updates = _report_updates;
      _report_updates.clear();
    }
  //WDEBUG_OUT << L"Unlocked reporting lock (thread: " << pthread_self() << L")" << std::endl;;
    
    while (!pending.empty())
      {
	Report *report( pending.front() );
	pending.pop_front();
	
	receive_report(report);
      }
    
    while (!updates.empty())
      {
	ReportUpdate *update( updates.front() );
	updates.pop_front();
	
	receive_report_update(update);
      }
  }
  
  return false;
}

void
Database:: receive_report(Report *report)
{
  _signal_report.emit(*report);
}

void
Database:: receive_report_update(ReportUpdate *update)
{
  _signal_report_update.emit(*update);
  
  if (update->get_status() == ReportUpdate::Ended)
    {
      Reports::iterator iter( std::find(_running_reports.begin(), _running_reports.end(), &update->get_report()) );
      assert(iter != _running_reports.end());
      Report *rep(*iter);
      _running_reports.erase(iter);
      delete rep;
    }
  
  delete update;
}

Database::SignalReport
Database:: signal_report()
{
  return _signal_report;
}

Database::SignalReportUpdate
Database:: signal_report_update()
{
  return _signal_report_update;
}


void
Database:: property_type_get_details (const Tk::String &domain,
				      const Tk::String &property_name,
				      Id *property_type_id,
				      Tk::String *property_value_type_name,
				      Tk::String *property_value_table_name)
{
  const static Tk::String
    sql(" select    pt.property_type_id"
	" ,         pvt.name"
	" ,         pvt.value_table"
	" from      property_type pt"
	" ,         property_value_type pvt"
	" where     pt.domain = ?"
	" and       pt.name = ?"
	" and       pvt.property_value_type_id = pt.property_value_type_id"
	);
  
  Connection::Mutex::Lock lock(_connection.get_mutex());
  PStatement *query( _connection.prepare_statement(sql) );
  
  query->setString(1, domain);
  query->setString(2, property_name);
  
  DataSet result( query->executeQuery() );
  
  bool valid( result->next() );
  if (!valid)
    throw UnknownType(property_name);
  
  if (property_type_id)
    *property_type_id = result->getLong(1);
  if (property_value_type_name)
    *property_value_type_name = result->getString(2);
  if (property_value_table_name)
    *property_value_table_name = result->getString(3);
}

Id
Database:: insert_property(const Id &property_type_id)
{
  Tk::Strings columns;
  columns.push_back("property_type_id");
  
  Connection::Mutex::Lock lock( _connection.get_mutex() );
  PStatementPtr stmt( _connection.get_table_insert_statement("property", columns) );
  
  stmt->setLong(1, property_type_id);
  stmt->execute();
  
  return _connection.get_last_insert_id();
}

void
Database:: insert_property_value_data (const Id &property_id,
				       const Tk::String &table_name,
				       Data::Value &value)
{
  Tk::Strings columns;
  columns.push_back("property_id");
  columns.push_back("property_value");

  Connection::Mutex::Lock lock( _connection.get_mutex() );
  PStatementPtr stmt( _connection.get_table_insert_statement(table_name, columns) );
  
  stmt->setLong(1, property_id);
  
  value.set_parameter(stmt.get(), 2);
  
  stmt->execute();
}

void
Database:: update_property_value_data (const Id &property_id,
				       const Tk::String &table_name,
				       Data::Value &value)
{
  std::wostringstream query;
  query << L" update " << table_name
	<< L" set    property_value = ?"
	<< L" where  property_id    = ?"
    ;

  Connection::Mutex::Lock lock( _connection.get_mutex() );
  PStatementPtr stmt( _connection.prepare_statement(query.str()) );
  
  value.set_parameter(stmt.get(), 1);
  stmt->setLong(2, property_id);
  
  stmt->execute();
}

void
Database:: insert_domain_property(const Tk::String &domain,
				  const Id &entity_id,
				  const Id &property_id)
{
  Tk::Strings columns;
  columns.push_back(domain + Tk::String("_id"));
  columns.push_back("property_id");
  
  Connection::Mutex::Lock lock( _connection.get_mutex() );
  PStatementPtr stmt( _connection.get_table_insert_statement(domain + Tk::String("_property"), columns) );
  
  stmt->setLong(1, entity_id);
  stmt->setLong(2, property_id);
  stmt->execute();
}


Id
Database:: set_property(const Tk::String &domain,
			const Id &entity_id,
			const Tk::String &property_name,
			Data::Value &value)
{
  //WDEBUG_OUT << L"Setting domain property '" << domain << L"." << property_name << "'.." << std::endl;
  Id property_type_id;
  Tk::String property_value_type_name;
  Tk::String property_value_table_name;
  
  //WDEBUG_OUT << L"Getting type details..." << std::endl;
  property_type_get_details (domain,
			     property_name,
			     &property_type_id,
			     &property_value_type_name,
			     &property_value_table_name);
  /*
  WDEBUG_OUT << L"...got type details (type name: '"
	     << property_value_type_name
	     << L"'; table name: '"
	     << property_value_table_name
	     << L"')"
	     << std::endl;
  */

  Id property_id;
  try
    {
      //WDEBUG_OUT << L"Finding property..." << std::endl;
      property_id = find_property(domain, entity_id, property_name);
      //WDEBUG_OUT << L"...found property (id: " << property_id << L")" << std::endl;
      
      //WDEBUG_OUT << L"Updating property value data..." << std::endl;
      update_property_value_data(property_id,
				 property_value_table_name,
				 value);
      //WDEBUG_OUT << L"...updated property value data" << std::endl;
    }
  catch (PropertyNotSet pns)
    {
      //WDEBUG_OUT << L"...property not set; inserting new property..." << std::endl;
      property_id = insert_property(property_type_id);
      //WDEBUG_OUT << L"...inserted property (id: " << property_id << L")" << std::endl;
      
      //WDEBUG_OUT << L"Inserting domain property..." << std::endl;
      insert_domain_property(domain, entity_id, property_id);
      //WDEBUG_OUT << L"...inserted domain property" << std::endl;
      
      //WDEBUG_OUT << L"Inserting property value data..." << std::endl;
      insert_property_value_data(property_id,
				 property_value_table_name,
				 value);
      //WDEBUG_OUT << L"...inserted property value data" << std::endl;
    }
  
  //WDEBUG_OUT << L"..set domain property '" << domain << L"." << property_name << L"'" << std::endl;
  return property_id;
}


Id
Database:: lookup_location(const Tk::String &name)
{
  const static Tk::String sql("select location_id from location where name = ?");
  Connection::Mutex::Lock lock( _connection.get_mutex() );
  PStatement *stmt( _connection.prepare_statement(sql) );
  stmt->setString(1, name);
  
  DataSet result( stmt->executeQuery() );
  
  bool valid( result->next() );
  if (!valid)
    throw std::runtime_error(std::string(_("Unknown location '")) + static_cast<std::string>(name) + _("'"));
  
  return result->getLong(1);
}

Id
Database:: insert_experiment(const Id &location_id)
{
  Tk::Strings columns;
  columns.push_back("location_id");
  
  Connection::Mutex::Lock lock( _connection.get_mutex() );
  PStatementPtr stmt( _connection.get_table_insert_statement("experiment", columns) );
  
  stmt->setLong(1, location_id);
  
  stmt->execute();
  
  return _connection.get_last_insert_id();
}

void
Database:: experiment_insert_operators(const Id &experiment_id, const Ids &operators, const Report *parent_report)
{
  Tk::Strings columns;
  columns.push_back("experiment_id");
  columns.push_back("operator_id");
  
  Connection::Mutex::Lock lock( _connection.get_mutex() );
  PStatementPtr stmt( _connection.get_table_insert_statement("experiment_operator", columns) );
  
  stmt->setLong(1, experiment_id);
  
  Report *rep( new Report("Storing experiment operators", parent_report) );
  send_report(rep);
  
  double op_count( operators.size() );
  unsigned op_done(0);
  for (Ids::const_iterator iter(operators.begin());
       iter != operators.end();
       ++iter)
    {
      stmt->setLong(2, *iter);
      stmt->execute();
      
      ++op_done;
      send_report_update(new ReportUpdate(*rep, ReportUpdate::Running, op_done/op_count));
    }
  
  send_report_update(new ReportUpdate(*rep, ReportUpdate::Ended));
}

void
Database:: experiment_insert_result(const Id &experiment_id,
				    const Tk::Resolver::Result &result,
				    const Report *parent_report)
{
  bool colours(result.front()[0].colour);
  
  Tk::Strings offset_columns;
  offset_columns.push_back("experiment_id");
  offset_columns.push_back("line_index");
  offset_columns.push_back("offset_index");
  
  Tk::StringsPtr colour_columns;
  if (colours)
    colour_columns.reset( new Tk::Strings(offset_columns) );
  
  offset_columns.push_back("offset");
  
  if (colours)
    {
      colour_columns->push_back("hue");
      colour_columns->push_back("saturation");
      colour_columns->push_back("value");
    }
  
  Connection::Mutex::Lock lock( _connection.get_mutex() );
  PStatementPtr offset_stmt( _connection.get_table_insert_statement("experiment_result", offset_columns) );
  
  PStatementPtr colour_stmt;
  if (colours)
    colour_stmt.reset( _connection.get_table_insert_statement("experiment_colour", *colour_columns) );
  
  offset_stmt->setLong(1, experiment_id);
  if (colours)
    colour_stmt->setLong(1, experiment_id);
  
  Report *rep( new Report("Storing experiment results", parent_report) );
  send_report(rep);
  
  unsigned ntraces( result.begin()->size() );
  unsigned line_index(0);
  double line_count( result.size() );
  unsigned offset_index;
  for (Tk::Resolver::Result::const_iterator iter(result.begin());
       iter != result.end();
       ++iter, ++line_index)
    {
      offset_stmt->setInt(2, line_index);
      if (colours)
	colour_stmt->setInt(2, line_index);
      
      for (offset_index = 0; offset_index < ntraces; ++offset_index)
	{
	  const Tk::Resolver::Offset &offset((*iter)[offset_index]);
	  
	  offset_stmt->setInt(3, offset_index);
	  offset_stmt->setInt(4, offset.offset);
	  offset_stmt->execute();
	  
	  if (colours)
	    {
	      colour_stmt->setInt(3, offset_index);
	      colour_stmt->setDouble(4, offset.colour->get_h());
	      colour_stmt->setDouble(5, offset.colour->get_s());
	      colour_stmt->setDouble(6, offset.colour->get_v());
	      colour_stmt->execute();
	    }
	}
      
      send_report_update(new ReportUpdate(*rep, ReportUpdate::Running, (line_index + 1)/line_count));
    }
  
  send_report_update(new ReportUpdate(*rep, ReportUpdate::Ended));
}

void
Database:: insert_experiment_run(ExperimentResult *experiment)
{
  WDEBUG_OUT << "Thread running..." << std::endl;

  Report *rep( new Report("Storing experiment in database") );
  send_report(rep);
  
  Report *subrep( new Report("Creating experiment", rep) );
  send_report(subrep);
  Id id( insert_experiment(experiment->get_location_id()) );
  send_report_update(new ReportUpdate(*subrep, ReportUpdate::Running, 1.0));
  send_report_update(new ReportUpdate(*subrep, ReportUpdate::Ended));
  send_report_update(new ReportUpdate(*rep, ReportUpdate::Running, (1.0/4.0)));
  
  subrep = new Report("Setting experiment properties", rep);
  send_report(subrep);
  Data::ValueMap *properties( experiment->take_properties() );
  double prop_count( properties->size() );
  unsigned prop_done(0);
  for (Data::ValueMap::iterator iter( properties->begin() );
       iter != properties->end();
       ++iter)
    {
      set_property("experiment", id, iter->first, iter->second);
      
      ++prop_done;
      send_report_update(new ReportUpdate(*subrep, ReportUpdate::Running, prop_done/prop_count));
    }
  send_report_update(new ReportUpdate(*subrep, ReportUpdate::Ended));
  send_report_update(new ReportUpdate(*rep, ReportUpdate::Running, (2.0/4.0)));
  
  
  experiment_insert_operators(id, experiment->get_operators(), rep);
  send_report_update(new ReportUpdate(*rep, ReportUpdate::Running, (3.0/4.0)));
  
  experiment_insert_result(id, experiment->get_result(), rep);
  send_report_update(new ReportUpdate(*rep, ReportUpdate::Running, (4.0/4.0)));
  
  send_report_update(new ReportUpdate(*rep, ReportUpdate::Ended));
  
  delete properties;
  delete experiment;
  
  WDEBUG_OUT << "...thread run" << std::endl;
}

void
Database:: insert_experiment(ExperimentResult *experiment)
{
  sigc::slot<void>
    slot( sigc::bind<ExperimentResult *>
	  ( sigc::mem_fun(*this, &Database::insert_experiment_run),
	    experiment ) );
  
  WDEBUG_OUT << "Creating thread..." << std::endl;
  Glib::Thread::create(slot, false);
  WDEBUG_OUT << "...created thread" << std::endl;
}

Id
Database:: choose_location()
{
  Tk::String
    sql(" select   e.location_id"
	" from     experiment e"
	" ,        experiment_property ep"
	" ,        property p"
	" ,        property_type pt"
	" ,        property_value_datetime pv"
	" where    ep.experiment_id    = e.experiment_id"
	" and      p.property_id       = ep.property_id"
	" and      pv.property_id      = p.property_id"
	" and      pt.property_type_id = p.property_type_id"
	" and      pt.domain           = ?"
	" and      pt.name             = ?"
	" order by pv.property_value desc"
	);
  
  Connection::Mutex::Lock lock( _connection.get_mutex() );
  PStatement *query( _connection.prepare_statement(sql) );
  query->setString(1, Tk::String("experiment"));
  query->setString(2, Tk::String("time stamp"));
  
  DataSetPtr result( new DataSet(query->executeQuery()) );
  
  bool valid( (*result)->next() );
  if (!valid)
    {
      sql = "select location_id from location";
      result.reset( _connection.execute_query(sql) );
      
      valid = (*result)->next();
    }
  
  if (!valid)
    throw std::runtime_error(_("Could not choose location"));

  return (*result)->getLong(1);
}

Id
Database:: choose_operator()
{
  Tk::String
    sql(" select   eo.operator_id"
	" from     experiment e"
	" ,        experiment_operator eo"
	" ,        experiment_property ep"
	" ,        property p"
	" ,        property_type pt"
	" ,        property_value_datetime pv"
	" where    eo.experiment_id    = e.experiment_id"
        " and      ep.experiment_id    = e.experiment_id"
	" and      p.property_id       = ep.property_id"
	" and      pv.property_id      = p.property_id"
	" and      pt.property_type_id = p.property_type_id"
	" and      pt.domain           = ?"
	" and      pt.name             = ?"
	" order by pv.property_value desc"
	);
  
  Connection::Mutex::Lock lock( _connection.get_mutex() );
  PStatement *query( _connection.prepare_statement(sql) );
  query->setString(1, Tk::String("experiment"));
  query->setString(2, Tk::String("time stamp"));
  
  DataSetPtr result( new DataSet(query->executeQuery()) );
  
  bool valid( (*result)->next() );
  if (!valid)
    {
      sql = "select operator_id from operator";
      result.reset( _connection.execute_query(sql) );
      
      valid = (*result)->next();
    }
  
  if (!valid)
    throw std::runtime_error(_("Could not choose operator"));

  return (*result)->getLong(1);
}

Code
Database:: choose_intent(unsigned trace_count, const Tk::String &algorithm_spec, bool test)
{
  const static Tk::String
    sql(" select   i.intent_code"
	" from     intent_domain id"
	" ,        intent i"
	" where    id.trace_count = ?"
	" and      i.intent_domain_id = id.intent_domain_id"
	" order by ("
	"            select   count(*)"
	"            from     property_value_string pvs"
	"            ,        property p"
	"            ,        property_type pt"
	"            ,        experiment_property ep"
	"            ,        experiment e"
	"            ,        experiment_property epa"
	"            ,        property pa"
	"            ,        property_type pta"
	"            ,        property_value_string pvsa"
	"            ,        experiment_property epts"
	"            ,        property pts"
	"            ,        property_type ptts"
	"            ,        property_value_string pvsts"
	"            where    pvs.property_value        = i.intent_code"
	"            and      p.property_id             = pvs.property_id"
	"            and      pt.property_type_id       = p.property_type_id"
	"            and      pt.domain                 = ?"
	"            and      pt.name                   = ?"
	"            and      ep.property_id            = p.property_id"
	"            and      e.experiment_id           = ep.experiment_id"
	"            and      epa.experiment_id         = e.experiment_id"
	"            and      pa.property_id            = epa.property_id"
	"            and      pta.domain                = ?"
	"            and      pta.name                  = ?"
	"            and      pvsa.property_id          = pa.property_id"
	"            and      pvsa.property_value       = ?"
	"            and      epts.experiment_id        = e.experiment_id"
	"            and      pts.property_id           = epts.property_id"
	"            and      ptts.domain               = ?"
	"            and      ptts.name                 = ?"
	"            and      pvsts.property_id         = pts.property_id"
	"            and      pvsts.property_value      = ?"
	"	   )"
	);
  
  Connection::Mutex::Lock lock( _connection.get_mutex() );
  PStatement* query( _connection.prepare_statement(sql) );
  
  query->setInt(1, static_cast<int>(trace_count));
  query->setString(2, Tk::String("experiment"));
  query->setString(3, Tk::String("intent code"));
  query->setString(4, Tk::String("experiment"));
  query->setString(5, Tk::String("algorithm spec"));
  query->setString(6, algorithm_spec);
  query->setString(7, Tk::String("experiment"));
  query->setString(8, Tk::String("test"));
  query->setBoolean(9, test);
  
  DataSet result( query->executeQuery() );
  
  bool valid( result->next() );
  if (!valid)
    throw std::runtime_error(_("Could not choose intent"));
  
  return result->getString(1);
}

IntentInfoList
Database:: get_intent_options(unsigned trace_count)
{
  const static Tk::String
    sql(" select   i.intent_code"
	" ,        i.description"
	" from     intent_domain id"
	" ,        intent i"
	" where    id.trace_count = ?"
	" and      i.intent_domain_id = id.intent_domain_id"
	);
  Connection::Mutex::Lock lock( _connection.get_mutex() );
  PStatement *query( _connection.prepare_statement(sql) );
  
  query->setInt(1, trace_count);
  
  DataSet result( query->executeQuery() );
  
  IntentInfoList info;
  for (result->next(); !result->isAfterLast(); result->next())
    info.push_back( IntentInfo(result->getString(1), result->getString(2)) );
  
  return info;
}

DataSet *
Database:: get_results(const Id &experiment_id)
{
  const static Tk::String
    sql(" select   offset"
	" from     experiment_result"
	" where    experiment_id = ?"
	" order by line_index"
	" ,        offset_index"
	);
  
  PStatement *stmt( _connection.prepare_statement(sql) );
  stmt->setLong(1, experiment_id);
  
  return new DataSet(stmt->executeQuery());
}

Id
Database:: find_property(const Tk::String &domain, const Id &entity_id, const Tk::String &property_name)
{
  Id property_type_id;
  
  property_type_get_details (domain,
			     property_name,
			     &property_type_id);
  
  std::wostringstream sqls;
  sqls << L" select   p.property_id"
       << L" from     " << domain << L" d"
       << L" ,        " << domain << L"_property dp"
       << L" ,        property p"
       << L" where    d." << domain << L"_id    = ?"
       << L" and      dp." << domain << L"_id   = d." << domain << L"_id"
       << L" and      p.property_id             = dp.property_id"
       << L" and      p.property_type_id        = ?"
    ;
  Tk::String sql(sqls.str());
  
  Connection::Mutex::Lock lock( _connection.get_mutex() );
  PStatement *query( _connection.prepare_statement(sql) );
  
  query->setLong(1, entity_id);
  query->setLong(2, property_type_id);
  
  DataSet result( query->executeQuery() );
  
  bool valid( result->next() );
  if (!valid)
    throw PropertyNotSet(domain, entity_id, property_name);
  
  return result->getLong(1);
}

Data::Value
Database:: get_property(const Tk::String &domain, const Id &entity_id, const Tk::String &property_name)
{
  Id property_type_id;
  Tk::String property_value_type_name;
  Tk::String property_value_table_name;
  
  property_type_get_details (domain,
			     property_name,
			     &property_type_id,
			     &property_value_type_name,
			     &property_value_table_name);
  
  std::wostringstream sqls;
  sqls << L" select   pv.property_value "
       << L" from     " << domain << L" d"
       << L" ,        " << domain << L"_property dp"
       << L" ,        property p"
       << L" ,        " << property_value_table_name << L" pv"
       << L" where    d." << domain << L"_id    = ?"
       << L" and      dp." << domain << L"_id   = d." << domain << L"_id"
       << L" and      p.property_id             = dp.property_id"
       << L" and      p.property_type_id        = ?"
       << L" and      pv.property_id            = p.property_id"
    ;
  Tk::String sql(sqls.str());
  
  Connection::Mutex::Lock lock( _connection.get_mutex() );
  PStatement *query( _connection.prepare_statement(sql) );
  
  query->setLong(1, entity_id);
  query->setLong(2, property_type_id);
  
  DataSet result( query->executeQuery() );
  
  bool valid( result->next() );
  if (!valid)
    throw PropertyNotSet(domain, entity_id, property_name);
  
  return Data::Value(result, 1);
}


Ids
Database:: get_ids(PStatement &stmt)
{
  WDEBUG_OUT << L"getting ids" << std::endl;
  
  WDEBUG_OUT << L"executing statement" << std::endl;
  DataSet result( stmt.executeQuery() );
  WDEBUG_OUT << L"executed statement" << std::endl;
  
  Ids ids;
  for (result->next(); !result->isAfterLast(); result->next())
    ids.push_back( result->getLong(1) );
  
  WDEBUG_OUT << L"returning ids" << std::endl;
  return ids;
}

Data::ValueMap
Database:: get_properties(const Tk::String &domain, const Id &entity_id)
{
  std::wostringstream sqls;
  sqls << L" select   pt.name "
       << L" from     " << domain << L" d"
       << L" ,        " << domain << L"_property dp"
       << L" ,        property p"
       << L" ,        property_type pt"
       << L" where    d." << domain << L"_id    = ?"
       << L" and      dp." << domain << L"_id   = d." << domain << L"_id"
       << L" and      p.property_id             = dp.property_id"
       << L" and      pt.property_type_id       = p.property_type_id"
    ;
  Tk::String sql(sqls.str());
  
  Tk::Strings names;
  {
    Connection::Mutex::Lock lock( _connection.get_mutex() );
    PStatement *query( _connection.prepare_statement(sql) );
    query->setLong(1, entity_id);
    
    DataSet result( query->executeQuery() );
    
    for (result->next(); !result->isAfterLast(); result->next())
      names.push_back( result->getString(1) );
  }
  
  Data::ValueMap properties;
  for (Tk::Strings::iterator iter( names.begin() );
       iter != names.end();
       ++iter)
    {
      try
	{
	  Data::Value value( get_property(domain, entity_id, *iter) );
	  properties.insert(Data::ValueMap::value_type(*iter, value));
	}
      catch (PropertyNotSet &pns) {}
    }
  
  return properties;
}

Data::TypeMap
Database:: get_domain_property_types(const Tk::String &domain)
{
  const static Tk::String
    sql(" select   pt.name as property_type_name"
	" ,        pvt.name as property_value_type_name"
	" from     property_type pt"
	" ,        property_value_type pvt"
	" where    pt.domain                  = ?"
	" and      pvt.property_value_type_id = pt.property_value_type_id"
	);
  Connection::Mutex::Lock lock( _connection.get_mutex() );
  PStatement *query( _connection.prepare_statement(sql) );
  
  query->setString(1, domain);
  
  DataSet result( query->executeQuery() );
  
  Data::TypeMap properties;
  for (result->next(); !result->isAfterLast(); result->next())
    properties[result->getString(1)] = Data::Value::get_type(result->getString(2));
  
  return properties;
}

Id
Database:: create_entity(const Tk::String &domain)
{
  Tk::Strings columns;
  columns.push_back(domain + Tk::String("_id"));
  
  Connection::Mutex::Lock lock( _connection.get_mutex() );
  PStatementPtr stmt( _connection.get_table_insert_statement(domain, columns) );
  
  stmt->execute();
  
  return _connection.get_last_insert_id();
}

Tk::Db::Ids
Database:: get_entities(const Tk::String &domain)
{
  std::wostringstream sqls;
  sqls << L" select   d." << domain << L"_id"
       << L" from     " << domain << L" d"
    ;
  Tk::String sql(sqls.str());
  
  Connection::Mutex::Lock lock( _connection.get_mutex() );
  DataSetPtr result( _connection.execute_query(sql) );
  
  Ids ids;
  for ((*result)->next(); !(*result)->isAfterLast(); (*result)->next())
    ids.push_back( (*result)->getLong(1) );
  
  return ids;
}
