/* -*- c++ -*-
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2008 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "libtkcommon/intl.h"
#include "libtkcommon/debug.h"

#include "libtkquery/identifier.h"
#include "libtkquery/parse_error.h"
#include "libtkdb/exception.h"

using namespace Tk::Query;

Identifier:: Identifier(Alpha *alpha1,
			const Location &alpha1_loc,
			Alpha *alpha2,
			const Location &alpha2_loc)
  : _domain( resolve_domain(alpha1_loc, alpha1->get_text()) ),
    _domain_loc( new Location(alpha1_loc) ),
    _property_name( alpha2->get_text() ),
    _property_loc( alpha2_loc )
{
}

Identifier:: Identifier(Alpha *alpha,
			const Location &alpha_loc,
			Name *name,
			const Location &name_loc)
  : _domain( resolve_domain(alpha_loc, alpha->get_text()) ),
    _domain_loc( new Location(alpha_loc) ),
    _property_name( name->get_name() ),
    _property_loc(name_loc)
{
}

Identifier:: Identifier(Name *name, const Location &name_loc)
  : _domain_loc(0),
    _property_name( name->get_name() ),
    _property_loc(name_loc)
{
}

Identifier:: Identifier(Alpha *alpha, const Location &alpha_loc)
  : _domain_loc(0),
    _property_name( alpha->get_text() ),
    _property_loc(alpha_loc)
{
}

Identifier:: ~Identifier()
{
  if (_domain_loc)
    delete _domain_loc;
}

Tk::String
Identifier::resolve_domain(const Location &loc,
			   const Tk::String &partial)
{
  Tk::Strings domains;
  domains.push_back("experiment");
  domains.push_back("operator");
  domains.push_back("location");
  
  for (Tk::Strings::iterator iter( domains.begin() );
       iter != domains.end();
       )
    {
      if (iter->size() < partial.size())
	iter = domains.erase(iter);
      else
	++iter;
    }
  
  for (unsigned i(0); i < partial.size(); ++i)
    {
      for (Tk::Strings::iterator iter( domains.begin() );
	   iter != domains.end();
	   )
	{
	  if ((*iter)[i] != partial[i])
	    iter = domains.erase(iter);
	  else
	    ++iter;
	}
      
      if (domains.size() < 2)
	break;
    }
  
  if (domains.size() != 1)
    throw SemanticError(loc,
			std::string(_("Ambiguous domain specifier `"))
			+ static_cast<std::string>(partial)
			+ _("'")
			);
  
  return *domains.begin();
}

Query *
Identifier:: build_query(Tk::Db::Database *database, const Tk::String &query_domain) const
{
  Tk::String id_domain;
  if (_domain.size() > 0)
    id_domain = _domain;
  else
    id_domain = query_domain;
  
  //WDEBUG_OUT << L"building Identifier (name: '" << _name << "')" << std::endl;
  Query *query( new Query(database) );
  
  TableAlias property( query->add_table("property") );
  TableAlias property_type( query->add_table("property_type") );
  TableAlias domain_property( query->add_table(id_domain + Tk::String("_property")) );
  
  TableAlias query_domain_alias( Query::get_table_abbrev(query_domain) );
  Tk::String query_domain_id( query_domain + Tk::String("_id") );
  
  const Tk::String property_id("property_id");
  const Tk::String property_type_id("property_type_id");
  
  if ("experiment" == query_domain)
    {
      if ("operator" == id_domain)
	{
	  const Tk::String operator_id("operator_id");
	  
	  TableAlias experiment_op( query->add_table("experiment_operator") );
	  TableAlias op( query->add_table("operator") );
	  
	  query->add_link_clause(experiment_op, query_domain_id, query_domain_alias, query_domain_id);
	  query->add_link_clause(op, operator_id, experiment_op, operator_id);
	  query->add_link_clause(domain_property, operator_id, op, operator_id);
	}
      else if ("location" == id_domain)
	{
	  const Tk::String location_id("location_id");
	  
	  TableAlias location( query->add_table("location") );
	  
	  query->add_link_clause(location, location_id, query_domain_alias, location_id);
	  query->add_link_clause(domain_property, location_id, location, location_id);
	}
      else
	{
	  query->add_link_clause(domain_property, query_domain_id, query_domain_alias, query_domain_id);
	}
    }
  else
    {
      if (id_domain != query_domain)
	throw SemanticError(*_domain_loc,
			    std::string("Invalid identifier domain `")
			    + static_cast<std::string>(id_domain)
			    + "' with query domain `"
			    + static_cast<std::string>(query_domain)
			    + "'"
			    );
      
      query->add_link_clause(domain_property, query_domain_id, query_domain_alias, query_domain_id);
    }
  
  TableAlias property_value;
  try
    {
      property_value = query->add_property_value_table(id_domain, _property_name);
    }
  catch (Tk::Db::UnknownType &ut)
    {
      throw SemanticError(_property_loc,
			  static_cast<std::string>("Unknown property `")
			  + static_cast<std::string>(_property_name)
			  + "' in domain `"
			  + static_cast<std::string>(id_domain)
			  + "'");
    }
  
  query->add_link_clause(property, property_id, domain_property, property_id);
  query->add_link_clause(property_value, property_id, property, property_id);
  query->add_link_clause(property_type, property_type_id, property, property_type_id);
  
  Tk::Db::Data::Values domain_params;
  domain_params.push_back(Tk::Db::Data::Value(Tk::Db::Data::String, static_cast<Tk::Db::Data::Datum>(id_domain)));
  query->add_link_clause(property_type + Tk::String(".domain = ?"), domain_params);
  
  Tk::Db::Data::Values name_params;
  name_params.push_back(Tk::Db::Data::Value(Tk::Db::Data::String, static_cast<Tk::Db::Data::Datum>(_property_name)));
  query->add_link_clause(property_type + Tk::String(".name = ?"), name_params);
  
  query->set_expr(property_value + Tk::String(".property_value"));
  
  return query;
}
