/*
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2004, 2005, 2006, 2007 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <wchar.h>

#include <iostream>
#include <string>

#include <SDL/SDL.h>

#include "libtkcommon/exceptions.h"
#include "libtkcommon/debug.h"
#include "libtkcommon/real_time.h"
#include "libtkrandom/types.h"
#include "resolver.h"
#include "types.h"

using namespace Tk::Resolver;

Resolver:: Resolver(const String &algo, unsigned bit_multiplier)
  : _algorithm(algo),
    _decision_tree(_algorithm.get_outcome_count(), bit_multiplier)
{
}

Result *
Resolver:: run(Tk::Random::Retriever &retriever,
	       Tk::Random::Randomness &randomness,
	       bool coloured,
	       unsigned &height,
	       unsigned &width,
	       unsigned trace_count,
	       const DirectionValue &arrow,
	       unsigned total_time,
	       double horizon,
	       bool test,
	       unsigned &retries,
	       Time &actual_time)
{
  Drawer drawer(_algorithm, coloured, height, width, trace_count, arrow, total_time, horizon, test);
  drawer.start();
  
  retries = _decision_tree.decide(retriever, randomness, drawer);
  
  Result *result( drawer.take_result() );
  if (result->empty())
    {
      delete result;
      throw Escaped();
    }
  
  height      = drawer.get_display_height();
  width       = drawer.get_display_width();
  actual_time = drawer.get_actual_time();
  
  return result;
}


Resolver::Drawer:: Drawer(const Algorithm &algorithm,
			  bool coloured,
			  unsigned height,
			  unsigned width,
			  unsigned trace_count, 
			  const DirectionValue &arrow,
			  unsigned total_time, 
			  double horizon,
			  bool test)
  : _algorithm(algorithm),
    _trace_count(trace_count),
    _test(test),
    _display(height, width, trace_count, arrow, total_time, horizon),
    _colours(0),
    _result( new Result )
{
  if (coloured)
    {
      _colours = new ColourGenerators;
      
      for (unsigned i(0); i < trace_count; ++i)
	_colours->push_back( ColourGenerator(_display.get_height()) );
    }
  
}

Resolver::Drawer:: ~Drawer()
{
  if (_colours)
    delete _colours;
  if (_result)
    delete _result;
}

void
Resolver::Drawer:: start()
{
  _display.count_down(5);
  _display.start();
}

bool
Resolver::Drawer:: affect(Tk::Random::Choice choice)
{
  if (SDL::check_escape(_test))
    {
      _result->clear();
      return false;
    }
  
  _choices.push_back(choice);
  
  if (_choices.size() < _trace_count)
    return true;
  
  Line l;
  unsigned i(0);
  for (Choices::iterator iter( _choices.begin() );
       iter != _choices.end();
       ++iter, ++i)
    {
      Colour *colour;
      if (_colours)
	colour = new Colour((*_colours)[i].get_colour());
      else
	colour = 0;
      
      l.push_back( Offset(_algorithm.choose_offset(*iter), colour) );
    }
  
  _display.draw_line(l);
  _actual_time = _display.stop();
  
  _choices.clear();
  _result->push_back(l);
  
  return _actual_time == 0;
}

unsigned
Resolver::Drawer:: get_display_height() const
{
  return _display.get_height();
}

unsigned
Resolver::Drawer:: get_display_width() const
{
  return _display.get_width();
}

const Time &
Resolver::Drawer:: get_actual_time() const
{
  return _actual_time;
}

Result *
Resolver::Drawer:: take_result()
{
  Result *r( _result );
  _result = 0;
  return r;
}
